@@ -1,3 +1,4 @@
+{-# LANGUAGE DeriveDataTypeable #-}
 {-
 Copyright (C) 2009 John MacFarlane <jgm@berkeley.edu>
 
@@ -23,17 +24,18 @@ module Text.TeXMath.Parser (expr, formula, Exp(..), TeXSymbolType(..), ArrayLine
 where
 
 import Control.Monad
-import Data.Char (isAlphaNum)
+import Data.Char (isAlphaNum, isDigit, isAscii)
 import qualified Data.Map as M
 import Text.ParserCombinators.Parsec
 import qualified Text.ParserCombinators.Parsec.Token as P
 import Text.ParserCombinators.Parsec.Language
+import Data.Generics
 
 data TeXSymbolType = Ord | Op | Bin | Rel | Open | Close | Pun | Accent
-                     deriving (Show, Read, Eq)
+                     deriving (Show, Read, Eq, Data, Typeable)
 
 data Alignment = AlignLeft | AlignCenter | AlignRight | AlignDefault
-                 deriving (Show, Read, Eq)
+                 deriving (Show, Read, Eq, Data, Typeable)
 
 type ArrayLine = [[Exp]]
 
@@ -51,12 +53,15 @@ data Exp =
   | EOver Exp Exp
   | EUnder Exp Exp
   | EUnderover Exp Exp Exp
+  | EUp Exp Exp
+  | EDown Exp Exp
+  | EDownup Exp Exp Exp
   | EUnary String Exp
   | EScaled String Exp
   | EStretchy Exp
   | EArray [Alignment] [ArrayLine]
   | EText String String
-  deriving (Show, Read, Eq)
+  deriving (Show, Read, Eq, Data, Typeable)
 
 texMathDef :: LanguageDef st
 texMathDef = LanguageDef 
@@ -89,6 +94,7 @@ expr1 =  choice [
   , array
   , diacritical
   , escaped
+  , unicode
   ]
 
 formula :: GenParser Char st [Exp]
@@ -99,11 +105,28 @@ formula = do
   return f
 
 expr :: GenParser Char st Exp
-expr = subSup <|> superOrSubscripted <|> expr1
+expr = do
+  a <- expr1
+  limits <- limitsIndicator
+  subSup limits a <|> superOrSubscripted limits a <|> return a
+
+limitsIndicator :: GenParser Char st (Maybe Bool)
+limitsIndicator =
+   try (symbol "\\limits" >> return (Just True))
+  <|> try (symbol "\\nolimits" >> return (Just False))
+  <|> return Nothing
 
 inbraces :: GenParser Char st Exp
 inbraces = liftM EGrouped (braces $ many $ notFollowedBy (char '}') >> expr)
 
+texToken :: GenParser Char st Exp
+texToken = inbraces <|> inbrackets <|>
+             do c <- anyChar
+                spaces
+                return $ if isDigit c
+                            then (ENumber [c])
+                            else (EIdentifier [c])
+
 inbrackets :: GenParser Char st Exp
 inbrackets = liftM EGrouped (brackets $ many $ notFollowedBy (char ']') >> expr)
 
@@ -118,16 +141,16 @@ basicEnclosure = choice $ map (\(s, v) -> try (symbol s) >> return v) enclosures
 
 left :: GenParser Char st Exp
 left = try $ do
-  try (symbol "\\left")
-  enc <- basicEnclosure <|> (symbol "." >> return (ESymbol Open "\xFEFF"))
+  symbol "\\left"
+  enc <- basicEnclosure <|> (try (symbol ".") >> return (ESymbol Open "\xFEFF"))
   case enc of
     (ESymbol Open _) -> tilRight enc <|> return (EStretchy enc)
     _ -> pzero
 
 right :: GenParser Char st Exp
 right = try $ do
-  try (symbol "\\right")
-  enc <- basicEnclosure <|> (symbol "." >> return (ESymbol Open "\xFEFF"))
+  symbol "\\right"
+  enc <- basicEnclosure <|> (try (symbol ".") >> return (ESymbol Close "\xFEFF"))
   case enc of
     (ESymbol Close x) -> return (EStretchy $ ESymbol Open x)
     _ -> pzero
@@ -149,12 +172,18 @@ scaledEnclosure = try $ do
        Just  r -> liftM (EScaled r . EStretchy) basicEnclosure
        Nothing -> pzero 
 
+endLine :: GenParser Char st Char
+endLine = try $ do
+  symbol "\\\\"
+  optional inbrackets  -- can contain e.g. [1.0in] for a line height, not yet supported
+  return '\n'
+
 arrayLine :: GenParser Char st ArrayLine
 arrayLine = notFollowedBy (try $ char '\\' >> symbol "end" >> return '\n') >>
-  sepBy1 (many (notFollowedBy (try $ char '\\' >> char '\\') >> expr)) (symbol "&")
+  sepBy1 (many (notFollowedBy endLine >> expr)) (symbol "&")
 
 array :: GenParser Char st Exp
-array = stdarray <|> eqnarray <|> cases <|> matrix
+array = stdarray <|> eqnarray <|> align <|> cases <|> matrix
 
 matrix :: GenParser Char st Exp
 matrix =  matrixWith "pmatrix" "(" ")"
@@ -167,24 +196,29 @@ matrixWith :: String -> String -> String -> GenParser Char st Exp
 matrixWith keywd opendelim closedelim =
   inEnvironment keywd $ do
     aligns <- option [] arrayAlignments
-    lines' <- sepEndBy1 arrayLine (try $ symbol "\\\\")
-    return $ EGrouped [ ESymbol Open opendelim
+    lines' <- sepEndBy1 arrayLine endLine
+    return $ EGrouped [ EStretchy (ESymbol Open opendelim)
                       , EArray aligns lines'
-                      , ESymbol Close closedelim]
+                      , EStretchy (ESymbol Close closedelim)]
 
 stdarray :: GenParser Char st Exp
 stdarray = inEnvironment "array" $ do
   aligns <- option [] arrayAlignments
-  liftM (EArray aligns) $ sepEndBy1 arrayLine (try $ symbol "\\\\")
+  liftM (EArray aligns) $ sepEndBy1 arrayLine endLine
 
 eqnarray :: GenParser Char st Exp
 eqnarray = inEnvironment "eqnarray" $
   liftM (EArray [AlignRight, AlignCenter, AlignLeft]) $
-    sepEndBy1 arrayLine (try $ symbol "\\\\")
+    sepEndBy1 arrayLine endLine
+
+align :: GenParser Char st Exp
+align = inEnvironment "align" $
+  liftM (EArray [AlignRight, AlignLeft]) $
+    sepEndBy1 arrayLine endLine
 
 cases :: GenParser Char st Exp
 cases = inEnvironment "cases" $ do
-  rs <- sepEndBy1 arrayLine (try $ symbol "\\\\")
+  rs <- sepEndBy1 arrayLine endLine
   return $ EGrouped [EStretchy (ESymbol Open "{"), EArray [] rs]
 
 arrayAlignments :: GenParser Char st [Alignment]
@@ -215,30 +249,54 @@ variable = do
   spaces
   return $ EIdentifier [v]
 
-subSup :: GenParser Char st Exp
-subSup = try $ do
-  a <- expr1
+isConvertible :: Exp -> Bool
+isConvertible (EMathOperator x) = x `elem` convertibleOps
+  where convertibleOps = ["lim","liminf","limsup","inf","sup"]
+isConvertible (ESymbol Rel _) = True
+isConvertible (ESymbol Bin _) = True
+isConvertible (EUnder _ _)    = True
+isConvertible (EOver _ _)     = True
+isConvertible (EUnderover _ _ _) = True
+isConvertible (ESymbol Op x) = x `elem` convertibleSyms
+  where convertibleSyms = ["\x2211","\x220F","\x22C2",
+           "\x22C3","\x22C0","\x22C1","\x2A05","\x2A06",
+           "\x2210","\x2A01","\x2A02","\x2A00","\x2A04"]
+isConvertible _ = False
+
+subSup :: Maybe Bool -> Exp -> GenParser Char st Exp
+subSup limits a = try $ do
   char '_'
   b <- expr1
   char '^'
   c <- expr
-  return $ ESubsup a b c 
+  return $ case limits of
+            Just True  -> EUnderover a b c
+            Nothing | isConvertible a -> EDownup a b c
+            _          -> ESubsup a b c
 
-superOrSubscripted :: GenParser Char st Exp
-superOrSubscripted = try $ do
-  a <- expr1
+superOrSubscripted :: Maybe Bool -> Exp -> GenParser Char st Exp
+superOrSubscripted limits a = try $ do
   c <- oneOf "^_"
   b <- expr
   case c of
-       '^' -> return $ ESuper a b
-       '_' -> return $ ESub a b
-       _   -> pzero 
+       '^' -> return $ case limits of
+                        Just True  -> EOver a b
+                        Nothing | isConvertible a -> EUp a b
+                        _          -> ESuper a b
+       '_' -> return $ case limits of
+                        Just True  -> EUnder a b
+                        Nothing | isConvertible a -> EDown a b
+                        _          -> ESub a b
+       _   -> pzero
 
 escaped :: GenParser Char st Exp
 escaped = lexeme $ try $ 
           char '\\' >>
           liftM (ESymbol Ord . (:[])) (satisfy $ not . isAlphaNum)
 
+unicode :: GenParser Char st Exp
+unicode = lexeme $ liftM (ESymbol Ord . (:[])) $ satisfy (not . isAscii)
+
 command :: GenParser Char st String
 command = try $ char '\\' >> liftM ('\\':) (identifier <|> lexeme (count 1 anyChar))
 
@@ -248,6 +306,7 @@ unaryOps = ["\\sqrt", "\\surd"]
 textOps :: M.Map String (String -> Exp)
 textOps = M.fromList
           [ ("\\textrm", EText "normal")
+          , ("\\mathrm", EText "normal")
           , ("\\text",   EText "normal")
           , ("\\mbox",   EText "normal")
           , ("\\mathbf", EText "bold")
@@ -257,8 +316,8 @@ textOps = M.fromList
           , ("\\mathtt", EText "monospace")
           , ("\\texttt", EText "monospace")
           , ("\\mathsf", EText "sans-serif")
-          , ("\\mathbb", EText "double-struck")
-          , ("\\mathcal", EText "script")
+          , ("\\mathbb", \e -> maybe (EText "double-struck" e) (ESymbol Pun) (M.lookup e mathbb))
+          , ("\\mathcal", \e -> maybe (EText "script" e) (ESymbol Pun) (M.lookup e mathcal))
           , ("\\mathfrak", EText "fraktur")
           ]
 
@@ -266,7 +325,7 @@ diacritical :: GenParser Char st Exp
 diacritical = try $ do
   c <- command
   case M.lookup c diacriticals of
-       Just r  -> liftM r inbraces
+       Just r  -> liftM r texToken
        Nothing -> pzero 
 
 diacriticals :: M.Map String (Exp -> Exp)
@@ -298,7 +357,7 @@ unary :: GenParser Char st Exp
 unary = try $ do
   c <- command
   unless (c `elem` unaryOps) pzero 
-  a <- inbraces
+  a <- texToken
   return $ EUnary c a
 
 text :: GenParser Char st Exp
@@ -313,15 +372,15 @@ root :: GenParser Char st Exp
 root = try $ do
   try (symbol "\\sqrt") <|> symbol "\\surd"
   a <- inbrackets
-  b <- inbraces
+  b <- texToken
   return $ EBinary "\\sqrt" b a
 
 binary :: GenParser Char st Exp
 binary = try $ do
   c <- command
   unless (c `elem` binaryOps) pzero 
-  a <- inbraces
-  b <- inbraces
+  a <- texToken
+  b <- texToken
   return $ EBinary c a b
 
 texSymbol :: GenParser Char st Exp
@@ -366,6 +425,14 @@ scalers = M.fromList
           , ("\\Bigg", "2.9")
           , ("\\big", "1.2")
           , ("\\Big", "1.6")
+          , ("\\biggr", "2.2")
+          , ("\\Biggr", "2.9")
+          , ("\\bigr", "1.2")
+          , ("\\Bigr", "1.6")
+          , ("\\biggl", "2.2")
+          , ("\\Biggl", "2.9")
+          , ("\\bigl", "1.2")
+          , ("\\Bigl", "1.6")
           ]
 
 enclosures :: [(String, Exp)]
@@ -381,8 +448,8 @@ enclosures = [ ("(", ESymbol Open "(")
              , ("\\rbrace", ESymbol Close "}")
              , ("\\llbracket", ESymbol Open "\x27E6")
              , ("\\rrbracket", ESymbol Close "\x230B")
-             , ("\\langle", ESymbol Open "\x3009")
-             , ("\\rangle", ESymbol Close "\x300A")
+             , ("\\langle", ESymbol Open "\x27E8")
+             , ("\\rangle", ESymbol Close "\x27E9")
              , ("\\lfloor", ESymbol Open "\x230A")
              , ("\\rfloor", ESymbol Close "\x230B")
              , ("\\lceil", ESymbol Open "\x2308")
@@ -591,6 +658,8 @@ symbols = M.fromList [
            , ("\\bigodot", ESymbol Op "\x2A00")
            , ("\\biguplus", ESymbol Op "\x2A04")
            , ("\\int", ESymbol Op "\x222B")
+           , ("\\iint", ESymbol Op "\x222C")
+           , ("\\iiint", ESymbol Op "\x222D")
            , ("\\oint", ESymbol Op "\x222E")
            , ("\\prime", ESymbol Ord "\x2032")
            , ("\\dots", ESymbol Ord "\x2026")
@@ -664,3 +733,130 @@ symbols = M.fromList [
            , ("\\tanh", EMathOperator "tanh")
            ] 
 
+-- MathML has a mathvariant attribute which is unimplemented in Firefox
+--    (see https://bugzilla.mozilla.org/show_bug.cgi?id=114365)
+-- Therefore, we translate mathcal to unicode symbols directly.
+-- This list is from http://www.w3.org/TR/MathML2/script.html
+mathcal :: M.Map String String
+mathcal = M.fromList [
+             ("A", "\x1D49C")
+           , ("B", "\x0212C")
+           , ("C", "\x1D49E")
+           , ("D", "\x1D49F")
+           , ("E", "\x02130")
+           , ("F", "\x02131")
+           , ("G", "\x1D4A2")
+           , ("H", "\x0210B")
+           , ("I", "\x02110")
+           , ("J", "\x1D4A5")
+           , ("K", "\x1D4A6")
+           , ("L", "\x02112")
+           , ("M", "\x02133")
+           , ("N", "\x1D4A9")
+           , ("O", "\x1D4AA")
+           , ("P", "\x1D4AB")
+           , ("Q", "\x1D4AC")
+           , ("R", "\x0211B")
+           , ("S", "\x1D4AE")
+           , ("T", "\x1D4AF")
+           , ("U", "\x1D4B0")
+           , ("V", "\x1D4B1")
+           , ("W", "\x1D4B2")
+           , ("X", "\x1D4B3")
+           , ("Y", "\x1D4B4")
+           , ("Z", "\x1D4B5")
+           , ("a", "\x1D4B6")
+           , ("b", "\x1D4B7")
+           , ("c", "\x1D4B8")
+           , ("d", "\x1D4B9")
+           , ("e", "\x0212F")
+           , ("f", "\x1D4BB")
+           , ("g", "\x0210A")
+           , ("h", "\x1D4BD")
+           , ("i", "\x1D4BE")
+           , ("j", "\x1D4BF")
+           , ("k", "\x1D4C0")
+           , ("l", "\x1D4C1")
+           , ("m", "\x1D4C2")
+           , ("n", "\x1D4C3")
+           , ("o", "\x02134")
+           , ("p", "\x1D4C5")
+           , ("q", "\x1D4C6")
+           , ("r", "\x1D4C7")
+           , ("s", "\x1D4C8")
+           , ("t", "\x1D4C9")
+           , ("u", "\x1D4CA")
+           , ("v", "\x1D4CB")
+           , ("w", "\x1D4CC")
+           , ("x", "\x1D4CD")
+           , ("y", "\x1D4CE")
+           , ("z", "\x1D4CF")
+           ]
+
+-- Similar to mathcal above, we translate manually.
+-- This list is from http://www.w3.org/TR/MathML2/double-struck.html
+mathbb :: M.Map String String
+mathbb = M.fromList [
+             ("A", "\x1D538")
+           , ("B", "\x1D539")
+           , ("C", "\x02102")
+           , ("D", "\x1D53B")
+           , ("E", "\x1D53C")
+           , ("F", "\x1D53D")
+           , ("G", "\x1D53E")
+           , ("H", "\x0210D")
+           , ("I", "\x1D540")
+           , ("J", "\x1D541")
+           , ("K", "\x1D542")
+           , ("L", "\x1D543")
+           , ("M", "\x1D544")
+           , ("N", "\x02115")
+           , ("O", "\x1D546")
+           , ("P", "\x02119")
+           , ("Q", "\x0211A")
+           , ("R", "\x0211D")
+           , ("S", "\x1D54A")
+           , ("T", "\x1D54B")
+           , ("U", "\x1D54C")
+           , ("V", "\x1D54D")
+           , ("W", "\x1D54E")
+           , ("X", "\x1D54F")
+           , ("Y", "\x1D550")
+           , ("Z", "\x02124")
+           , ("a", "\x1D552")
+           , ("b", "\x1D553")
+           , ("c", "\x1D554")
+           , ("d", "\x1D555")
+           , ("e", "\x1D556")
+           , ("f", "\x1D557")
+           , ("g", "\x1D558")
+           , ("h", "\x1D559")
+           , ("i", "\x1D55A")
+           , ("j", "\x1D55B")
+           , ("k", "\x1D55C")
+           , ("l", "\x1D55D")
+           , ("m", "\x1D55E")
+           , ("n", "\x1D55F")
+           , ("o", "\x1D560")
+           , ("p", "\x1D561")
+           , ("q", "\x1D562")
+           , ("r", "\x1D563")
+           , ("s", "\x1D564")
+           , ("t", "\x1D565")
+           , ("u", "\x1D566")
+           , ("v", "\x1D567")
+           , ("w", "\x1D568")
+           , ("x", "\x1D569")
+           , ("y", "\x1D56A")
+           , ("z", "\x1D56B")
+           , ("0", "\x1D7D8")
+           , ("1", "\x1D7D9")
+           , ("2", "\x1D7DA")
+           , ("3", "\x1D7DB")
+           , ("4", "\x1D7DC")
+           , ("5", "\x1D7DD")
+           , ("6", "\x1D7DE")
+           , ("7", "\x1D7DF")
+           , ("8", "\x1D7E0")
+           , ("9", "\x1D7E1")
+           ]
