@@ -5,10 +5,11 @@ import Text.TeXMath
 import Data.Maybe (fromMaybe)
 import Control.Monad
 import Text.JSON
+import Codec.Binary.UTF8.String (decodeString)
 
 cgiMain :: CGI CGIResult
 cgiMain = do
-  latexFormula <- liftM (fromMaybe "") $ getInput "latexFormula"
+  latexFormula <- liftM (decodeString . fromMaybe "") $ getInput "latexFormula"
   setHeader "Content-type" "text/xhtml; charset=UTF-8"
   output . encodeStrict $
     case texMathToMathML DisplayBlock latexFormula of
